#!/usr/bin/env python

import os, sys, datetime, logging

from kafka.admin import KafkaAdminClient
from argparse import ArgumentParser

from kafka.errors import (UnknownError, KafkaConnectionError, FailedPayloadsError,
                          KafkaTimeoutError, KafkaUnavailableError,
                          LeaderNotAvailableError, UnknownTopicOrPartitionError,
                          NotLeaderForPartitionError, ReplicaNotAvailableError)


def _exit(message = None, rc=0):
  if (message): printMsg(message)
  printMsg("Exiting [rc={}]".format(rc))
  sys.exit(rc)


def timeFormat():
  return '%b.%d.%Y-%H.%M.%S.%f'


def printMsg(message):
  print ("[{}] [{}] {}".format(datetime.datetime.now().strftime(timeFormat()), GNode, message))


class MyKafkaAdmin:
    kafka_service = os.environ.get('KAFKA_SERVER')
    kafka_port = os.environ.get('KAFKA_PORT')

    if not kafka_service: kafka_service = "localhost"
    if not kafka_port: kafka_port = "9092"

    server = kafka_service + ':' + kafka_port

    def __init__(self, topic, rtopic):

      # Topic:
      printMsg("Working topic: '{}'".format(topic))

      # Action 'remove topic':
      printMsg("Remove topic from message bus: {}".format(str(rtopic)))

      if (rtopic):
        admin = KafkaAdminClient(bootstrap_servers=self.server)
        try:
          admin.delete_topics([topic])
        except UnknownTopicOrPartitionError as e:
          printMsg("UnknownTopicOrPartitionError => missing messages for topic (removal considered successful)")


def parse_arguments():

  parser = ArgumentParser(description='Kafka Manager')
  parser.add_argument('-n', '--node', help='Node name (spaces are removed)', required=True)
  parser.add_argument('-t', '--topic', help='Kafka topic', required=True)
  parser.add_argument('-r', '--remove-topic', help='Remove topic from message bus', required=False, action='store_true')
  parser.add_argument('-d', '--debug', help='Debug kafka', required=False, action='store_true')

  arguments = parser.parse_args()

  return arguments


#####################
#      M A I N      #
#####################

if __name__ == "__main__":

  arguments = parse_arguments()

  GNode = arguments.node.replace(" ", "")
  topic = arguments.topic
  rtopic = arguments.remove_topic
  debug = arguments.debug

  printMsg("Starting")

  if (debug):
    now = datetime.datetime.now().strftime(timeFormat())
    logfile='{}-Admin-{}.log'.format(GNode, now)
    printMsg("Logging kafka in: '{}'".format(logfile))
    logging.basicConfig(filename=logfile, format='%(asctime)s - %(message)s', datefmt=timeFormat(), level=logging.DEBUG)

  try:
    admin = MyKafkaAdmin(topic, rtopic)
    _exit()

  except Exception as e:
    _exit(e, 1)
