// ANNA - Anna is Not Nothingness Anymore                                                         //
//                                                                                                //
// (c) Copyright 2005-2015 Eduardo Ramos Testillano & Francisco Ruiz Rayo                         //
//                                                                                                //
// See project site at http://redmine.teslayout.com/projects/anna-suite                           //
// See accompanying file LICENSE or copy at http://www.teslayout.com/projects/public/anna.LICENSE //


/**
   Realiza peticiones automaticas sobre el servidor HTTP de operaciones aritmeticas.

   La cadencia de envio de mensajes se establece mediante un temporizador.
   Los operadores se calculan de forma aleatoria.

   El servidor de este cliente: http_server.p o http_rserver.p
*/

#include <iostream>

#include <string.h>
#include <unistd.h>

#include <anna/core/core.hpp>
#include <anna/app/functions.hpp>
#include <anna/comm/comm.hpp>

#include <anna/xml/DocumentMemory.hpp>
#include <anna/xml/Node.hpp>

#include <anna/timex/Engine.hpp>
#include <anna/timex/Clock.hpp>

#include <anna/http/Request.hpp>
#include <anna/http/Response.hpp>
#include <anna/http/Transport.hpp>
#include <anna/http/Handler.hpp>

#include <anna/http/wims20/ClientSide.hpp>

static const Millisecond Resolution(250);
static const Millisecond Period(500);
static const Millisecond OneSecond(1000);

class Sender : public anna::timex::Clock {
public:
   Sender () : Clock ("Sender", Period), 
      a_messageByTick (0), 
      a_nquarter (0), 
      a_errorCounter (0),
      a_txMessageCounter (0),
      a_request (NULL)
   {
      a_httpRequest.setMethod (http::Method::Type::Get);
   }

   void setMessageBySecond (const int messageBySecond) throw () { a_messageByTick = messageBySecond / (OneSecond / Period); }
   void setRequest (http::wims20::ClientSide* request) throw () { a_request = request; }

   int getTxMessageCounter () const throw () { return a_txMessageCounter; }

private:
   int a_messageByTick;
   int a_nquarter;
   int a_errorCounter;
   int a_txMessageCounter;
   http::Request a_httpRequest;
   http::wims20::ClientSide* a_request;

   /* Se invoca 4 veces por segundo */
   bool tick () throw (RuntimeException);
};

class MyHandler : public http::Handler {
public:
   MyHandler () : http::Handler ("http_client::MyHandler") {;}

private:
   xml::DocumentMemory a_xmlResponse;
   
   void evRequest (ClientSocket&, const http::Request&) throw (RuntimeException) {;}
   void evResponse (ClientSocket&, const http::Response&) throw (RuntimeException);
   
//   static bool isOk (const test::Response& response) throw ();
};

class MyCommunicator : public comm::Communicator {
public:
   MyCommunicator () : comm::Communicator (), a_avgResponseTime (0), a_rxMessageCounter (0) {;}

   void count (const Millisecond delay) throw (RuntimeException);
   
   // Sustituye la redefinicin de los siguientes mtodos
   void eventBreakConnection (const comm::ClientSocket&) throw ();
   using comm::Communicator::eventBreakConnection;

private:
   int a_avgResponseTime;
   int a_rxMessageCounter;
   MyHandler a_httpHandler;
   
   void eventReceiveMessage (ClientSocket&, const Message&) throw (RuntimeException);
};

class HeavyWIMS20Client : public anna::comm::Application {
public:
   HeavyWIMS20Client ();

   Server* getServer () const throw () { return a_server; }
   const Sender* getSender () const throw () { return &a_sender; }

private:
   MyCommunicator a_communicator;
   anna::timex::Engine a_timeController;
   Sender a_sender;
   Server* a_server;
   http::wims20::ClientSide* a_request;

   void initialize () throw (RuntimeException);
   void run () throw (RuntimeException);
};

using namespace std;

int main (int argc, const char** argv)
{
   CommandLine& commandLine (CommandLine::instantiate ());
   HeavyWIMS20Client app;

   srand (time (NULL));

   try {
      commandLine.initialize (argv, argc);
      commandLine.verify ();

      Logger::setLevel (Logger::Information);
      string traceFile ("client.");
      traceFile += anna::functions::asString ((int) getpid ());
      traceFile += ".trace";
      Logger::initialize ("http_client", new TraceWriter (traceFile.c_str (),4096000));

      app.start ();
   }
   catch (Exception& ex) {
      cout << ex.asString () << endl;
   }

   return 0;
}

HeavyWIMS20Client::HeavyWIMS20Client () :
   Application ("http_client", "Cliente HTTP", "1.0"),
   a_communicator (),
   a_timeController (OneSecond, Resolution),
   a_request (NULL)
{
   CommandLine& commandLine (CommandLine::instantiate ());

   commandLine.add ("p", CommandLine::Argument::Mandatory, "Puerto en el que el servidor atiende respuestas.");
   commandLine.add ("a", CommandLine::Argument::Mandatory, "Direccin IP Puerto en el que el servidor atiende respuestas.");
   commandLine.add ("n", CommandLine::Argument::Mandatory, "Numero de mensajes por segundo");
   commandLine.add ("domain", CommandLine::Argument::Mandatory, "http://domain-openapis/path-openapis/serviceID/guid/other_possible_levels?query_parameters");
   commandLine.add ("path", CommandLine::Argument::Optional, "http://domain-openapis/path-openapis/serviceID/guid/other_possible_levels?query_parameters");
   commandLine.add ("service", CommandLine::Argument::Mandatory, "http://domain-openapis/path-openapis/serviceID/guid/other_possible_levels?query_parameters");
   commandLine.add ("guid", CommandLine::Argument::Mandatory, "http://domain-openapis/path-openapis/serviceID/guid/other_possible_levels?query_parameters");
   commandLine.add ("other", CommandLine::Argument::Optional, "http://domain-openapis/path-openapis/serviceID/guid/other_possible_levels?query_parameters");
   commandLine.add ("trace", CommandLine::Argument::Optional, "Nivel de trazas (debug,warning, error,...)");
}

void HeavyWIMS20Client::initialize ()
   throw (RuntimeException)
{
   CommandLine& cl (CommandLine::instantiate ());

   Network& network = Network::instantiate ();

   a_server = network.createServer (cl.getValue ("a"), cl.getIntegerValue ("p"), true, &http::Transport::getFactory ());
   a_sender.setMessageBySecond (cl.getIntegerValue ("n"));

   if (cl.exists ("trace"))
      Logger::setLevel (Logger::asLevel (cl.getValue ("trace")));

   const char* domain = cl.getValue ("domain");

   if (cl.exists ("path")) 
      a_request = new http::wims20::ClientSide (domain, cl.getValue ("path"));
   else
      a_request = new http::wims20::ClientSide (domain);

   a_request->setServiceID (cl.getValue ("service"));
   a_request->setGUID (cl.getValue ("guid"));

   if (cl.exists ("other")) 
      a_request->addOtherLevel (cl.getValue ("other"));
   
   a_sender.setRequest (a_request);
}

void HeavyWIMS20Client::run ()
   throw (RuntimeException)
{
   a_timeController.activate (a_sender);

   a_communicator.accept ();
}

void MyCommunicator::eventReceiveMessage (ClientSocket& clientSocket, const Message& message)
   throw (RuntimeException)
{
   LOGMETHOD (TraceMethod tm ("MyCommunicator", "eventReceiveMessage", ANNA_FILE_LOCATION));

   if (clientSocket.support (http::Transport::className ()) == false)
      return;

   a_httpHandler.apply (clientSocket, message);
}

void MyCommunicator::count (const Millisecond delay)
   throw (RuntimeException)
{
   Guard guard (this, "MyCommunicator::count");
   
   a_rxMessageCounter ++;
   a_avgResponseTime += delay;
}

void MyCommunicator::eventBreakConnection (const ClientSocket& clientSocket)
   throw ()
{
   if (a_rxMessageCounter == 0) {
      LOGWARNING (
         string msg ("MyCommunicator::eventBreakConnection | ");
         msg += clientSocket.asString ();
         Logger::warning (msg, ANNA_FILE_LOCATION);
      );
      return;
   }

   LOGNOTICE (
      HeavyWIMS20Client& app = static_cast <HeavyWIMS20Client&> (anna::app::functions::getApp ());   
      string msg ("Tiempo medio respuesta: ");
      msg += anna::functions::asString (a_avgResponseTime / a_rxMessageCounter);
      msg += " ms";
      msg += anna::functions::asText (" | Rx: ", a_rxMessageCounter);
      msg += anna::functions::asText (" | Tx: ", app.getSender ()->getTxMessageCounter ());
      Logger::notice (msg, ANNA_FILE_LOCATION);

      cout << msg << endl << endl;
   );
   requestStop ();
   comm::Communicator::eventBreakConnection (clientSocket);
}

bool Sender::tick ()
   throw (RuntimeException)
{
   Server* server = static_cast <HeavyWIMS20Client&> (anna::app::functions::getApp ()).getServer ();
   Communicator* communicator = anna::app::functions::component <Communicator> (ANNA_FILE_LOCATION);

   if (a_errorCounter > 50) {
      communicator->requestStop ();
      Logger::warning ("Terminado por errores continuos en la conexion", ANNA_FILE_LOCATION);
      return false;
   }

   for (int n = 0; n < a_messageByTick && communicator->hasRequestedStop () == false; n ++) {
      a_request->setParameter ("Operator", "a");
      a_request->setParameter ("ValueOne", rand () % 1000);
      a_request->setParameter ("ValueTwo", rand () % 1000);
      a_request->setParameter ("Time", anna::functions::millisecond ());

      try {         
         a_request->codeOn (a_httpRequest);
         server->send (a_httpRequest);
         a_txMessageCounter ++;
      }
      catch (RuntimeException& ex) {
         a_errorCounter ++;
         ex.trace ();
         break;
      }
   }

   return true;
}

void MyHandler::evResponse (ClientSocket& clientSocket, const http::Response& response)
   throw (RuntimeException)
{
   TraceMethod tm ("MyHandler", "evResponse", ANNA_FILE_LOCATION);

   if (response.getStatusCode () != 200) 
      return;
   
   // El servidor implementado en Java retornar un 500 cuando alcanze el n mximo de mensajes.
   if (response.getStatusCode () == 500) {
      LOGWARNING (
         string msg (" MyHandler::evResponse  | ");
         msg += response.asString ();
         msg += " | Servidor java notifica fin de la prueba.";
         Logger::debug (msg, ANNA_FILE_LOCATION);
      );      
      eventBreakConnection (clientSocket);
      return;      
   }
      
   a_xmlResponse.initialize (response.getBody ());

   const xml::Node* root = a_xmlResponse.parse ();
   
   const anna::Millisecond now = anna::functions::millisecond ();
   const anna::Millisecond past(root->getAttribute ("Time")->getIntegerValue ()); 
   
   const int delay = now - past; 

   if (delay < 0) {
      LOGWARNING (
         string msg = a_xmlResponse.getContentAsCString ();
         msg += anna::functions::asText (" | Delay: ", delay);
         Logger::warning (msg, ANNA_FILE_LOCATION);
      );
      return;
   }

   app::functions::component <MyCommunicator> (ANNA_FILE_LOCATION)->count (Millisecond(delay));
}
