// ANNA - Anna is Not Nothingness Anymore                                                         //
//                                                                                                //
// (c) Copyright 2005-2015 Eduardo Ramos Testillano & Francisco Ruiz Rayo                         //
//                                                                                                //
// See project site at http://redmine.teslayout.com/projects/anna-suite                           //
// See accompanying file LICENSE or copy at http://www.teslayout.com/projects/public/anna.LICENSE //


#ifndef anna_comm_Codec_hpp
#define anna_comm_Codec_hpp

#include <anna/comm/CompatCodec.hpp>

namespace anna {

class Second;
class Millisecond;
class Microsecond;
namespace comm {

/**
   Codificador/Decodificador compatible para transporte de datos

   Esta clase ofrece una forma muy eficaz para codificar estructuras de datos complejas y pretende
   sustituir los tipicos aplanadores y desaplanadores ya que permite que cualquier proceso
   defina facilmente la forma de transferir cierta informacion sobre un bloque de memoria
   y viceversa.

   Vamos a ver un ejemplo de uso. Primero vamos a definir la clase MensajeDemo:

* \code
*
* #include <anna.comm.Codec.h>
*
* class MensajeDemo : public Codec {
* public:
*    static const Codec::Type type = 10;
*
*    MensajeDemo () : Codec (type),
*       a_dataBlock (true)
*    {
*       attach ("Entero", a_entero);
*       attach ("Cadena",  a_cadena);
*       attach ("Bloque datos",  a_dataBlock);
*    }
*
*    // Accesores
*    const int obtenerEntero () const { return a_entero; }
*    const std::string& obtenerCadena () const { return a_cadena; }
*    const DataBlock& obtenerDataBlock () const { return a_dataBlock; }
*
*    // Modificadores
*    void establecerEntero (const int entero) { a_entero = entero; }
*    void establecerCadena (const std::string& cadena) { a_cadena = cadena; }
*    void establecerCadena (const char* cadena) { a_cadena = cadena; }
*    void establecerDataBlock (const DataBlock& dataBlock) { a_dataBlock = dataBlock; }
*
* private:
*    int a_entero;
*    std::string a_cadena;
*    DataBlock a_dataBlock;
* };
*
* \endcode
*
   La clase que use esta clase para enviar un mensaje debe establecer los valores de cada uno de
   los datos mediante los modificadores definidos y luego invocar al metodo #code que
   transferiria el contenido de las variables asociadas a este mensaje a un bloque de memoria,
   que normalmente seriausado como mensaje.

   Por otra parte la clase que recibe el mensaje invocar�al metodo #decode que transfiere
   el contenido del bloque de memoria a cada una de las variables asociadas al mensaje.
   Posteriormente, podremos acceder al contenido de cada una de las variables asociadas al
   mensaje atraves de los accesores definidos para el caso.

   Estos codificadores nos dan la posibilidad de definir variables opcionales, de forma que
   una determinada variable puede no ser transferida al bloque de memoria, y por tanto puede
   no ser recibida en el otro extremo. Ver el metodo #isNull y #setNull para mas informacion.

   \warning Esta clase no establece proteccion ante accesos concurrentes
*/
class Codec : public CompatCodec {
public:
  /**
     Constructor.

     @param type Tipo por el que sera conocido este tipo de mensaje.
     @param scramble Indica si el mensaje debe ser codificado de forma que no se pueda ver el contenido
     del mismo con una simple herramienta de monitorizacion de mensajes de red. Por defecto esta
     activo ya que la codificacion realizada es muy simple y rapida y el tiempo empleado es casi
     inapreciable.
  */
  explicit Codec(const Type type, const bool scramble = true) : CompatCodec(type, scramble) {;}

  /**
     Asocia el valor recibido como parametro al dato interno identificado por `id'. Esta clase
     estapensada principalmente para comunicar entre si procesos remotos
     La unica restriccion que se les imponen a ambos que es compartan el significado que cada uno
     de ellos le da a un determinado `id'.

     \param name Nombre logico de la variable
     @param value Rerencia a una variable de nuestro entorno usada para transferir los datos de la memoria
     al mensaje y viceversa.
     @return Un puntero que hace referencia al nuevo dato interno que ha sido creado.
  */
  const Variable* attach(const char* name, std::string& value) throw(RuntimeException) { return CompatCodec::attach(name, size(), value); }

  /**
     Asocia el valor recibido como parametro al dato interno identificado por `id'. Esta clase
     estapensada principalmente para comunicar entre si procesos remotos
     La unica restriccion que se les imponen a ambos que es compartan el significado que cada uno
     de ellos le da a un determinado `id'.

     \param name Nombre logico de la variable
     @param value Rerencia a una variable de nuestro entorno usada para transferir los datos de la memoria
     al mensaje y viceversa.
     @return Un puntero que hace referencia al nuevo dato interno que ha sido creado.
  */
  const Variable* attach(const char* name, const char*& value) throw(RuntimeException) { return CompatCodec::attach(name, size(), value); }

  /**
     Asocia el valor recibido como parametro al dato interno identificado por `id'. Esta clase
     estapensada principalmente para comunicar entre si procesos remotos.
     La unica restriccion que se les imponen a ambos que es compartan el significado que cada uno
     de ellos le da a un determinado `id'.

     \param name Nombre logico de la variable
     @param value Rerencia a una variable de nuestro entorno usada para transferir los datos de la memoria
     al mensaje y viceversa.
     @return Un puntero que hace referencia al nuevo dato interno que ha sido creado.
  */
  const Variable* attach(const char* name, int& value) throw(RuntimeException) { return CompatCodec::attach(name, size(), value); }

  /**
     Asocia el valor recibido como parametro al dato interno identificado por `id'. Esta clase
     estapensada principalmente para comunicar entre si procesos remotos.
     La unica restriccion que se les imponen a ambos que es compartan el significado que cada uno
     de ellos le da a un determinado `id'.

     \param name Nombre logico de la variable
     @param value Rerencia a una variable de nuestro entorno usada para transferir los datos de la memoria
     al mensaje y viceversa.
     @return Un puntero que hace referencia al nuevo dato interno que ha sido creado.
  */
  const Variable* attach(const char* name, S64& value) throw(RuntimeException) { return CompatCodec::attach(name, size(), value); }

  /**
     Asocia el valor recibido como parametro al dato interno identificado por `id'. Esta clase
     estapensada principalmente para comunicar entre si procesos remotos.
     La unica restriccion que se les imponen a ambos que es compartan el significado que cada uno
     de ellos le da a un determinado `id'.

     \param name Nombre logico de la variable
     @param value Rerencia a una variable de nuestro entorno usada para transferir los datos de la memoria
     al mensaje y viceversa.
     @return Un puntero que hace referencia al nuevo dato interno que ha sido creado.
  */
  const Variable* attach(const char* name, bool& value) throw(RuntimeException) { return CompatCodec::attach(name, size(), value); }

  /**
     Asocia el valor recibido como parametro al dato interno identificado por `id'. Esta clase
     estapensada principalmente para comunicar entre si procesos remotos.
     La unica restriccion que se les imponen a ambos que es compartan el significado que cada uno
     de ellos le da a un determinado `id'.

     \param name Nombre logico de la variable
     @param value Rerencia a una variable de nuestro entorno usada para transferir los datos de la memoria
     al mensaje y viceversa. Debe tener activado el sistema de copia profunda.
     @return Un puntero que hace referencia al nuevo dato interno que ha sido creado.
  */
  const Variable* attach(const char* name, DataBlock& value) throw(RuntimeException) { return CompatCodec::attach(name, size(), value); }

  /**
     Asocia el valor recibido como parametro al dato interno identificado por `id'. Esta clase
     estapensada principalmente para comunicar entre si procesos remotos.
     La unica restriccion que se les imponen a ambos que es compartan el significado que cada uno
     de ellos le da a un determinado `id'.

     \param name Nombre logico de la variable
     @param value Rerencia a una variable de nuestro entorno usada para transferir los datos de la memoria
     al mensaje y viceversa. Debe tener activado el sistema de copia profunda.
     @return Un puntero que hace referencia al nuevo dato interno que ha sido creado.
  */
  const Variable* attach(const char* name, float& value) throw(RuntimeException) { return CompatCodec::attach(name, size(), value); }

  /**
     Asocia el valor recibido como parametro al dato interno identificado por `id'. Esta clase
     estapensada principalmente para comunicar entre si procesos remotos.
     La unica restriccion que se les imponen a ambos que es compartan el significado que cada uno
     de ellos le da a un determinado `id'.

     \param name Nombre logico de la variable
     @param value Rerencia a una variable de nuestro entorno usada para transferir los datos de la memoria
     al mensaje y viceversa. Debe tener activado el sistema de copia profunda.
     @return Un puntero que hace referencia al nuevo dato interno que ha sido creado.
  */
  const Variable* attach(const char* name, double& value) throw(RuntimeException) { return CompatCodec::attach(name, size(), value); }

  /**
     Asocia el valor recibido como parametro al dato interno identificado por `id'. Esta clase
     estapensada principalmente para comunicar entre si procesos remotos.
     La unica restriccion que se les imponen a ambos que es compartan el significado que cada uno
     de ellos le da a un determinado `id'.

     \param name Nombre logico de la variable
     @param value Rerencia a una variable de nuestro entorno usada para transferir los datos de la memoria
     al mensaje y viceversa. Debe tener activado el sistema de copia profunda.
     @return Un puntero que hace referencia al nuevo dato interno que ha sido creado.
  */
  const Variable* attach(const char* name, Second& value) throw(RuntimeException);

  /**
     Asocia el valor recibido como parametro al dato interno identificado por `id'. Esta clase
     estapensada principalmente para comunicar entre si procesos remotos.
     La unica restriccion que se les imponen a ambos que es compartan el significado que cada uno
     de ellos le da a un determinado `id'.

     \param name Nombre logico de la variable
     @param value Rerencia a una variable de nuestro entorno usada para transferir los datos de la memoria
     al mensaje y viceversa. Debe tener activado el sistema de copia profunda.
     @return Un puntero que hace referencia al nuevo dato interno que ha sido creado.
  */
  const Variable* attach(const char* name, Millisecond& value) throw(RuntimeException);

  /**
     Asocia el valor recibido como parametro al dato interno identificado por `id'. Esta clase
     estapensada principalmente para comunicar entre si procesos remotos
     La unica restriccion que se les imponen a ambos que es compartan el significado que cada uno
     de ellos le da a un determinado `id'.

     \param name Nombre logico de la variable
     @param value Rerencia a una variable de nuestro entorno usada para transferir los datos de la memoria
     al mensaje y viceversa. Debe tener activado el sistema de copia profunda.
     @return Un puntero que hace referencia al nuevo dato interno que ha sido creado.
  */
  const Variable* attach(const char* name, Microsecond& value) throw(RuntimeException);
};

}
}

#endif
