// ANNA - Anna is Not Nothingness Anymore                                                         //
//                                                                                                //
// (c) Copyright 2005-2015 Eduardo Ramos Testillano & Francisco Ruiz Rayo                         //
//                                                                                                //
// See project site at http://redmine.teslayout.com/projects/anna-suite                           //
// See accompanying file LICENSE or copy at http://www.teslayout.com/projects/public/anna.LICENSE //


#ifndef anna_comm_Delivery_hpp
#define anna_comm_Delivery_hpp

#include <vector>

#include <anna/core/mt/Mutex.hpp>
#include <anna/config/defines.hpp>

namespace anna {

namespace xml {
class Node;
}

namespace comm {

class Cluster;
class Resource;

/**
   Clase que modela un algoritmo de reparto de carga.
*/
class Delivery  : public Mutex {
public:
  typedef std::vector <Resource*> Resources;
  typedef Resources::iterator iterator;
  typedef Resources::const_iterator const_iterator;

  /*
   * Establece el periodo de comprobacion de recuperacion de los
   * recursos en los que se han detectado problemas internos.
   */
  static const Millisecond DefaultRecoveryTime;

  /**
     Destructor.
  */
  virtual ~Delivery() { a_resources.clear(); }

  /**
     Devuelve el nombre del reparto. Coincide con el indicado en el constructor.
     \return el nombre del reparto. Coincide con el indicado en el constructor.
  */
  const std::string& getName() const throw() { return a_name; }

  /**
     Inicializa el reparto de recursos.
  */
  void initialize() throw(RuntimeException);

  /**
     Devuelve el estado anterior que tenia esta instancia. El resultado de #isAvailable
     puede ser distinto a este metodo ya que a diferencia de este se basa en el estado
     actual de los procesos asociados a este servicio, mientras que este metodo devuelve
     el estado en el que estaban la ultima vez que se hizo una comprobacion.
  */
  bool wasAvailable() const throw() { return a_isAvailable; }

  /*
   * Obtiene el periodo de comprobacion de recuperacion de los recusos que han sido
   * desactivados debido a que se han detectado problemas internos.
   *
   * \return El periodo de comprobacion de recuperacion
   */
  const Millisecond &getRecoveryTime() const throw() { return a_recoveryTime; }

  /**
   * Establece el periodo de comprobacion de recuperacion de los recusos que han sido
   * desactivados debido a que se han detectado problemas internos.
   *
   * \param recoveryTime Periodo de comprobacion de recuperacion.
   */
  void setRecoveryTime(const Millisecond &recoveryTime) throw() { a_recoveryTime = recoveryTime; }

  /**
     Comprueba la lista de recursos remotos para comprobar si hay alguno que este disponible
     (ver Resource::isAvailable) y no este deshabilitado (ver Resource::disable).

     \return \em true si tiene algun recurso remoto utilizado o \em false en otro caso.
  */
  virtual bool isAvailable() const throw();

  /**
     Devuelve la instancia del recurso remoto que debemos usar en esta ocasion.
     @return La instancia del recurso remoto que debemos usar en esta ocasion.
     \warning Antes de invocar a este mtodo la instancia se debe proteger de accesos concurrentes
  */
  Resource* apply() throw(RuntimeException);

  /**
     Este metodo sirve a nuestra aplicacion para notificar que el recurso ha dejado de
     estar disponible.
     \param resource Recurso remoto en el que hemos detectado el error.
     \return \em true si ninguno de los recursos asociados a este reparto de carga estan disponibles o \em false en otro caso.
     \warning Antes de invocar a este mtodo la instancia se debe proteger de accesos concurrentes
  */
  bool fault(const Resource* resource) throw();

  /**
     Este metodo sirve a nuestra aplicacion para que el recurso vuelve a estar disponible.
     \param resource Recurso remoto que hemos recuperado.
     \return \em true si alguno de los recursos asociados a este reparto de carga pasa a estar disponible o \em false en otro caso.
     \warning Antes de invocar a este mtodo la instancia se debe proteger de accesos concurrentes
  */
  bool recover(const Resource* resource) throw();

  /**
     Indica si reparto contiene la referencia a un determinado recurso.
     \return \em true si contiene la referencia al recurso recibido como parametro o \em false
     en otro caso.
  */
  bool contains(const Resource* resource) const throw();

  /**
     Devuelve un iterador al comienzo de la lista de recursos remotos asociados a este reparto.
     \return Un iterador al comienzo de la lista de recursos remotos asociados a este reparto.
  */
  iterator begin() throw() { return a_resources.begin(); }

  /**
     Devuelve un iterador al comienzo de la lista de recursos remotos asociados a este reparto.
     \return Un iterador al comienzo de la lista de recursos remotos asociados a este reparto.
  */
  const_iterator begin() const throw() { return a_resources.begin(); }

  /**
     Devuelve un iterador al final de la lista de recursos remotos asociados a este reparto.
     \return Un iterador al final de la lista de recursos remotos asociados a este reparto.
  */
  iterator end() throw() { return a_resources.end(); }

  /**
     Devuelve el numero de recursos asociados a la lista.
     \return El numero de recursos asociados a la lista.
  */
  int size() const throw() { return a_resources.size(); }

  /**
     Devuelve un iterador al final de la lista de recursos remotos asociados a este reparto.
     \return Un iterador al final de la lista de recursos remotos asociados a este reparto.
  */
  const_iterator end() const throw() { return a_resources.end(); }

  /**
     Devuelve una cadena con la informacion referente a este objeto.
     @return Una cadena con la informacion referente a este objeto.
  */
  virtual std::string asString() const throw();

  /**
     Devuelve un nodo XML con la informacion referente a este objeto.
     \param parent Nodo XML a partir del cual introducir la informacion.
     \return Un nodo XML con la informacion referente a este objeto.
  */
  virtual xml::Node* asXML(xml::Node* parent) const throw();

  /**
     Devuelve el recurso remoto apuntado por el iterador recibido como parametro.
     \param ii Iterador usado para recorrer los recursos asociados a este reparto.
     \return El recurso remoto apuntado por el iterador recibido como parametro.
  */
  static Resource* resource(iterator& ii) throw() { return *ii; }

  /**
     Devuelve el recurso remoto apuntado por el iterador recibido como parametro.
     \param ii Iterador usado para recorrer los recursos asociados a este reparto.
     \return El recurso remoto apuntado por el iterador recibido como parametro.
  */
  static const Resource* resource(const_iterator& ii) throw() { return *ii; }

protected:
  /**
     Constructor.
     @param name Nombre logico de este reparto.
  */
  Delivery(const char* name) :
    a_name(name),
    a_isAvailable(true),
    a_recoveryTime(DefaultRecoveryTime),
    a_timeStamp(0) {;}

  /**
     Conecta el recurso remoto recibido como parametro con este reparto.
     @param resource Instancia del recurso que vamos a registrar en este reparto.
  */
  void add(Resource* resource) throw(RuntimeException);

  /**
     Inicializa el reparto de recursos.
  */
  virtual void do_initialize() throw(RuntimeException) = 0;

  /**
     Devuelve la instancia del recurso remoto que debemos que debemos usar en esta ocasion.
     @return La instancia del recurso remoto que debemos usar en esta ocasion.
  */
  virtual Resource* do_apply() throw(RuntimeException) = 0;

  /**
     Este metodo sirve a nuestra aplicacion para indicar que no fue posible usar el
     recurso obtenido mediante #apply.
     \param resource Recurso remoto que ha ocasionado el error.
     \return \em true si ninguno de los recursos asociados a este reparto de carga estan disponibles o \em false en otro caso.
  */
  virtual bool do_fault(const Resource* resource) throw();

  /**
     Este metodo sirve a nuestra aplicacion para indicar que ha recuperado la conexion
     con el recurso.
     \param resource Recurso remoto con el que ha recuperado la conexion.
  */
  virtual bool do_recover(const Resource* resource) throw();

  /**
     Este metodo sirve a nuestra aplicacion para indicar que ha recuperado la conexion
     con el recurso.
     \param resource Recurso remoto con el que ha recuperado la conexion.
  */
  virtual bool do_contains(const Resource* resource) const throw();

private:
  std::string a_name;
  Resources a_resources;
  bool a_isAvailable;
  Millisecond a_recoveryTime;
  Millisecond a_timeStamp;

  bool unsafe_recover(const Resource* resource) throw();

  static bool internalErrorDetected(const Resource* resource) throw();
};

}
}

#endif
