// ANNA - Anna is Not Nothingness Anymore                                                         //
//                                                                                                //
// (c) Copyright 2005-2015 Eduardo Ramos Testillano & Francisco Ruiz Rayo                         //
//                                                                                                //
// See project site at http://redmine.teslayout.com/projects/anna-suite                           //
// See accompanying file LICENSE or copy at http://www.teslayout.com/projects/public/anna.LICENSE //


#ifndef anna_core_util_Configuration_hpp
#define anna_core_util_Configuration_hpp

#include <stdlib.h>

#include <vector>
#include <map>
#include <string>

#include <anna/core/functions.hpp>
#include <anna/core/util/Variable.hpp>
#include <anna/core/RuntimeException.hpp>

namespace anna {

/**
   Clase para recoger parametros de un determinado archivo de configuracion.
*/
class Configuration {
public:
  static const char* defaultSection;

  /**
     Constructor.
  */
  Configuration() {;}

  /**
  * Destructor.
  */
  ~Configuration() { removeAll(); }

  /**
     Carga en memoria el archivo indicado como argumento.

     @param configFile Ruta completa con el nombre del archivo de configuracion a cargar.
     Cualquier otro archivo procesado anteriormente con esta instancia se perdera.
  */
  void load(const char* configFile) throw(RuntimeException);

  /**
    Establece el valor por defecto para una determinada variable, es decir, en caso de que
    la variable no exista en el fichero de configuracion cargado (ver #load) devolvera
    el valor establecido mediante este metodo.

    @param sectionName Nombre de la seccion a la que pertenece la variable.
    @param variableName Nombre de la variable de configuracion.
    @param defaultValue Valor por defecto de la variable. ste valor solo sera devuelto en caso
    de que la variable indicada por la seccion y el nombre de variable no este contenido en
    el archivo de configuracion cargado.
  */
  void setDefaultValue(const char* sectionName, const char* variableName, const char* defaultValue)
  throw(RuntimeException);

  /**
     Devuelve el valor asociada a la variable indicada.

    @param sectionName Nombre de la seccion a la que pertenece la variable.
    @param variableName Nombre de la variable de configuracion.
    @param strict Si es true indica que debe devolver el valor estricto de la variable, de forma,
    que si esta variable no esta contenida en el archivo de configuracion aun habiendo asociado
    un valor por defecto (ver #setDefaultValue) devolvera NULL.

    @return El valor asociado a la variable. Puede ser NULL.
  */
  const char* getValue(const char* sectionName, const char* variableName, const bool strict = false) const
  throw(RuntimeException);

  /**
     Devuelve el valor asociada a la variable indicada.
    @param sectionName Nombre de la seccion a la que pertenece la variable.
    @param variableName Nombre de la variable de configuracion.
    @param strict Si es true indica que debe devolver el valor estricto de la variable, de forma,
    que si esta variable no esta contenida en el archivo de configuracion aun habiendo asociado
    un valor por defecto (ver #setDefaultValue) devolvera NULL.

    @return El valor asociado a la variable.
  */
  int getIntegerValue(const char* sectionName, const char* variableName, const bool strict = false) const
  throw(RuntimeException);

  /**
     Devuelve el estado de existencia o no de la variable indicada.
    @param sectionName Nombre de la seccion a la que pertenece la variable.
    @param variableName Nombre de la variable de configuracion.

    @return true si la variable existe, y false en otro caso. Solo deberia invocarse despues de
    invocar al metodo #load.
  */
  bool exists(const char* sectionName, const char* variableName) const throw();

  /**
     Devuelve la cadena por la que podemos buscar el componente.
     \return La cadena por la que podemos buscar el componente.
     \see Application::find
  */
  static const char* getClassName() { return "anna::Configuration"; }

private:
  class VariableEx : public Variable {
  public:
    typedef std::vector <VariableEx*> Vector;

    VariableEx(const char* variableName) :
      Variable(variableName, Variable::Type::String),
      a_defaultValue(NULL) {}

    void setDefaultValue(const char* defaultValue) { a_defaultValue = defaultValue; }

    const char* getDefaultValue() const { return a_defaultValue; }

  private:
    const char* a_defaultValue;
  };

  std::map <std::string, VariableEx::Vector*> a_sections;

  Configuration(const Configuration& other);  // No implementado

  void initialize() throw(RuntimeException) {;}
  void stop() throw() {;}

  void removeAll() throw();
  bool processSection(const int nline, char* buffer, std::string& currentSection);
  void processVariable(const int nline, char* buffer, const std::string& currentSection) throw(RuntimeException);
  VariableEx* createVariable(const std::string& section, const char* variableName) throw();
  VariableEx* find(const std::string& section, const char* variableName) throw();
  const VariableEx* find(const std::string& section, const char* variableName) const
  throw() {
    return const_cast <Configuration*>(this)->find(section, variableName);
  }

  static char* strip(char* buffer);
};

}

#endif

