// ANNA - Anna is Not Nothingness Anymore                                                         //
//                                                                                                //
// (c) Copyright 2005-2015 Eduardo Ramos Testillano & Francisco Ruiz Rayo                         //
//                                                                                                //
// See project site at http://redmine.teslayout.com/projects/anna-suite                           //
// See accompanying file LICENSE or copy at http://www.teslayout.com/projects/public/anna.LICENSE //


#ifndef anna_core_util_SortedVector_hpp
#define anna_core_util_SortedVector_hpp

#include <map>

#include <anna/core/RuntimeException.hpp>
#include <anna/config/defines.hpp>

namespace anna {

/**
   Patron para ordenar instancias de objetos en base de una clave.

   \param T Clase del patron.
   \param SortBy Clase que ofrece el valor por el que ordenar. Debe implementar un metodo constante con la
    signatura: TKey value (const T*)
   \param TKey Tipo de clave usado para calcular la ordenacion. Debe implementar los operadores '=', '<' y '==' y el
   contructor copia.

   \warning no actua como clase segura en MT, ver #anna::SafeSortedVector.
*/
template < typename T, typename SortBy, typename TKey = int > class SortedVector : public std::map <TKey, T*> {
public:
  typedef typename std::map <TKey, T*> container;
  typedef typename container::iterator iterator;
  typedef typename container::const_iterator const_iterator;
  typedef typename container::value_type value_type;

  /**
     Constructor.
  */
  SortedVector() {;}

  /**
     Constructor copia.
     \param other Instancia de la que copiar.
  */
  explicit SortedVector(const SortedVector& other) :
    container(other) {}

  /**
     Destructor.
  */
  virtual ~SortedVector() { container::clear(); }

  /**
     Devolvera \em true si la instancia recibida como parametro esta contenido en el
     map o \em en otro caso. Si la instancia recibida es NULL siempre devolvera \em false.
     \param t Instancia a comprobar.
     \return \em true si la instancia recibida como parametro esta contenido en el
     map o \em en otro caso.
  */
  bool contains(const T* t) const
  throw() {
    if(t == NULL)
      return false;

    TKey key(SortBy::value(t));
    return find(key) != NULL;
  }

  /**
     Incorpora la instancia recibida como parametro en la lista ordenada de objetos.
     \param t Instancia a guardar en el map. Si es NULL la operacion no tendra ningun efecto.
     \return \em true si ha registrado la nueva instancia o \em false en otro caso.
  */
  bool add(T* t)
  throw(RuntimeException) {
    if(t == NULL)
      return false;

    TKey key(SortBy::value(t));
    value_type v(key, t);
    std::pair <iterator, bool> result = container::insert(v);
    return result.second;
  }

  /**
     Elimina la instancia recibida como parametro de la lista ordenada de objetos.
     \param t Instancia a guardar en el map. Si es NULL la operacion no tendra ningun efecto.
     \return \em true si ha eliminado la instancia o \em false en otro caso.
  */
  bool erase(T* t)
  throw(RuntimeException) {
    if(t == NULL)
      return false;

    bool result = false;
    TKey key(SortBy::value(t));
    iterator ii = container::find(key);

    if(ii != container::end()) {
      container::erase(ii);
      result = true;
    }

    return result;
  }

  /**
     Devuelve la instancia asociada a la clave recibida como parametro o NULL si no existe.
     \param key Clave a buscar en el map.
     \return la instancia asociada a la clave recibida como parametro o NULL si no existe.
  */
  T* find(const TKey key)
  throw() {
    iterator ii = container::find(key);
    return (ii == container::end()) ? NULL : ii->second;
  }

  /**
     Devuelve la instancia asociada a la clave recibida como parametro o NULL si no existe.
     \param key Clave a buscar en el map.
     \return la instancia asociada a la clave recibida como parametro o NULL si no existe.
  */
  const T* find(const TKey key) const throw() {
    return const_cast <SortedVector <T, SortBy, TKey>*>(this)->find(key);
  }

  /**
     Devuelve el objeto referenciado por el iterador.
     \return El objeto referenciado por el iterador.
  */
  static T* data(iterator ii) throw() { return ii->second; }

  /**
     Devuelve el objeto referenciado por el iterador.
     \return El objeto referenciado por el iterador.
  */
  static const T* data(const_iterator ii) throw() { return ii->second; }
};

}

#endif
