// ANNA - Anna is Not Nothingness Anymore                                                         //
//                                                                                                //
// (c) Copyright 2005-2015 Eduardo Ramos Testillano & Francisco Ruiz Rayo                         //
//                                                                                                //
// See project site at http://redmine.teslayout.com/projects/anna-suite                           //
// See accompanying file LICENSE or copy at http://www.teslayout.com/projects/public/anna.LICENSE //


#ifndef anna_diameter_codec_basetypes_Address_hpp
#define anna_diameter_codec_basetypes_Address_hpp


// Local
#include <anna/diameter/codec/basetypes/OctetString.hpp>

#include <anna/config/defines.hpp>


namespace anna {

namespace diameter {

namespace codec {

class Avp;

namespace basetypes {

/**
* Diameter Address container
*
* Diameter address includes 2 bytes with ip version and 4/16 more for IPv4 and IPv6 respectively
*
* <pre>
*    IPv6 ADDRESS FORMAT CONSIDERATIONS
*
*      IPv6 addresses have two logical parts: a 64-bit network prefix, and a 64-bit host address part.
*      The host address is often automatically generated from the interface MAC address.
*      An IPv6 address is represented by 8 groups of 16-bit hexadecimal values separated by colons (:) shown as follows:
*
*      A typical example of an IPv6 address is
*
*          2001:0db8:85a3:0000:0000:8a2e:0370:7334.
*
*      The hexadecimal digits are case-insensitive.
*
*      The 128-bit IPv6 address can be abbreviated with the following rules:
*
*          * Rule one: Leading zeroes within a 16-bit value may be omitted. For example, the address fe80:0000:0000:0000:0202:b3ff:fe1e:8329
*                      may be written as fe80:0:0:0:202:b3ff:fe1e:8329
*
*          * Rule two: A single occurrence of consecutive groups of zeroes within an address may be replaced by a double colon.
*                      For example, fe80:0:0:0:202:b3ff:fe1e:8329 becomes fe80::202:b3ff:fe1e:8329
*
*
*    RULE ONE IS AUTOMATICALLY APPLIED BY sscanf and %X parsing
*    RULE TWO IS NOT APPLIED OVER THIS CLASS METHODS
*
*
*      In environments of dual-stack hosts that support both IPv4 and IPv6, an IPv4 address is expressed as an IPv6 address in an IPv6-mapped address.
*      For example, the IPv4-mapped IPv6 address ::ffff:c000:280 is usually written as ::ffff:192.0.2.128, thus expressing clearly the original IPv4
*      address that was mapped to IPv6.
*
*      The address ::ffff:1.2.3.4 is IPv4-mapped, different than address ::1.2.3.4 which is IPV4-compatible.
* </pre>
*/
class Address : OctetString {

  iana_address_t a_address;
  bool a_abbreviatePresentation; // specially for IPv6 address type


  // Only for derived diameter type:
  void updateBasic() throw(anna::RuntimeException);

  void setPrintableString(const char * printableString) throw(anna::RuntimeException);


public:

  /**
  * Default constructor
  */
  Address() : a_abbreviatePresentation(true) {};

  /**
  * Destructor
  */
  virtual ~Address() {;}
 
  /**
  * Presentation mode: abbreviate (i.e. two standard rules applied for IPv6) or expanded (i.e. groups of zeroes on IP address)
  *
  * @param abb Abbreviate mode boolean indicator
  */
  void setAbbreviatePresentation(bool abb) throw() { a_abbreviatePresentation = abb; };



  // Class-specific ////////////////////////////////////////////////////////////////////////////////////
  //
  /**
  * Gets the IANA address
  *
  * @return IANA address
  */
  const iana_address_t& getIANAAddress() const throw() { return a_address; }

  /**
  * Sets the IANA address
  *
  * @param address IANA address
  */
  void setIANAAddress(const iana_address_t& address) throw(anna::RuntimeException) { a_address = address; updateBasic(); }
  //
  //////////////////////////////////////////////////////////////////////////////////////////////////////


  // gets

  std::string getFormatName() const throw() { return "Address"; }


  // helpers

  /**
  * Gets the natural/smart string representation for IANA address
  * Launch exception when data is not printable: only IPv4 and IPv6 are printable-supported
  *
  * @return Natural/smart string representation for IANA address (only IPv4 and IPv6 are printable-supported)
  */
  std::string asPrintableString() throw(anna::RuntimeException);


  std::string asString() throw(anna::RuntimeException) { return a_address.asString(); }


  // sets

  void decode(const char* buffer, const int size) throw(anna::RuntimeException);


  // exports /////////////////////////////
  using AvpData::getSize;
  using AvpData::code;
  //using OctetString::asPrintableString;
  using AvpData::asDataBlockString;
  using OctetString::asString;
  using AvpData::asHexString;
  //using OctetString::decode;
  using AvpData::fromPrintableString;
  using AvpData::fromHexString;
};

}
}
}
}

#endif
