// ANNA - Anna is Not Nothingness Anymore                                                         //
//                                                                                                //
// (c) Copyright 2005-2015 Eduardo Ramos Testillano & Francisco Ruiz Rayo                         //
//                                                                                                //
// See project site at http://redmine.teslayout.com/projects/anna-suite                           //
// See accompanying file LICENSE or copy at http://www.teslayout.com/projects/public/anna.LICENSE //


#include <unistd.h>

#include <anna/core/tracing/Logger.hpp>
#include <anna/core/RuntimeException.hpp>
#include <anna/core/DataBlock.hpp>
#include <anna/core/tracing/TraceLogger.hpp>
#include <anna/core/functions.hpp>

using namespace anna;

Logger::Writer* Logger::st_writer = NULL;
bool Logger::st_enabled(true);

#ifndef _MT
pid_t Logger::st_pid(-1);
#endif

#ifdef _DEBUG
Logger::Level Logger::st_level(Logger::Debug);
#else
Logger::Level Logger::st_level(Logger::Warning);
#endif

NRMutex Logger::st_mutex;

void Logger::initialize(const char* ident)
throw() {
  if(st_writer == NULL) {
    st_writer = new TraceLogger;
    st_writer->initialize(ident);
  }
}

void Logger::initialize(const char* ident, Writer* writer)
throw() {
  if(st_writer == NULL && writer != NULL) {
    writer->initialize(ident);
    st_writer = writer;
  }
}

void Logger::showPID(const bool show)
throw() {
#ifndef _MT

  if(show == true) {
    if(st_pid == -1)
      st_pid = getpid();
  } else
    st_pid = -1;

#endif
}

void Logger::write(const Level level, const char* text, const char* fromFile, const int fromLine)
throw() {
#ifndef _MT

  if(isActive(level) && st_writer != NULL) {
    if(st_pid == -1)
      st_writer->do_write(level, "%s (%d) | %s", fromFile, fromLine, text);
    else
      st_writer->do_write(level, "%s (%d) | pid: %d | %s", fromFile, fromLine, st_pid, text);
  }

#else

  if(isActive(level) && st_writer != NULL)
    st_writer->do_write(level, "%s (%d) | thr: 0x%x | %s", fromFile, fromLine, (unsigned int) pthread_self(), text);

#endif
}

void Logger::write(const Level level, const char* text, const char* value, const char* fromFile, const int fromLine)
throw() {
#ifndef _MT

  if(isActive(level) && st_writer != NULL) {
    if(st_pid == -1)
      st_writer->do_write(level, "%s (%d) | %s | %s", fromFile, fromLine, text, value);
    else
      st_writer->do_write(level, "%s (%d) | pid: %d | %s | %s", fromFile, fromLine, st_pid, text, value);
  }

#else

  if(isActive(level) && st_writer != NULL)
    st_writer->do_write(level, "%s (%d) | thr: 0x%x | %s | %s", fromFile, fromLine, (unsigned int) pthread_self(), text, value);

#endif
}

void Logger::write(const Level level, const char* text, const int value, const char* fromFile, const int fromLine)
throw() {
#ifndef _MT

  if(isActive(level) && st_writer != NULL) {
    if(st_pid == -1)
      st_writer->do_write(level, "%s (%d) | %s | %d (%x)", fromFile, fromLine, text, value, value);
    else
      st_writer->do_write(level, "%s (%d) | pid: %d | %s | %d (%x)", fromFile, fromLine, st_pid, text, value, value);
  }

#else

  if(isActive(level) && st_writer != NULL)
    st_writer->do_write(level, "%s (%d) | thr: 0x%x | %s | %d (%x)", fromFile, fromLine, (unsigned int) pthread_self(), text, value, value);

#endif
}

void Logger::write(const Level level, const char* text, const DataBlock& value, const char* fromFile, const int fromLine)
throw() {
  if(isActive(level) && st_writer != NULL) {
#ifndef _MT

    try {
      st_writer->do_write(level, "%s (%d) | %s | Data block: %s", fromFile, fromLine, text, functions::asString(value).c_str());
    } catch(...) {
      st_writer->do_write(level, "%s (%d) | %s | Data block: <fault!>", fromFile, fromLine, text);
    }

#else

    try {
      st_writer->do_write(level, "%s (%d) | thr: 0x%x | %s | Data block: %s", fromFile, fromLine, (unsigned int) pthread_self(), text, functions::asString(value).c_str());
    } catch(...) {
      st_writer->do_write(level, "%s (%d) | thr: 0x%x | %s | Data block: <fault!>", fromFile, fromLine, (unsigned int) pthread_self(), text);
    }

#endif
  }
}

void Logger::disable()
throw(RuntimeException) {
  st_mutex.lock();
  st_enabled = false;
  st_mutex.unlock();
}

void Logger::enable()
throw(RuntimeException) {
  st_mutex.lock();
  st_enabled = true;
  st_mutex.unlock();
}

Logger::Level Logger::asLevel(const char* stringLevel)
throw(RuntimeException) {
  static struct {
    const char* stringLevel;
    const Level level;
  } values [] = {
    { "emergency", Emergency }, { "alert", Alert }, { "critical", Critical }, { "error", Error },
    { "warning", Warning }, { "notice", Notice }, { "information", Information }, { "debug", Debug },
    { "local0", Local0 }, { "local1", Local1 }, { "local2", Local2 }, { "local3", Local3 },
    { "local4", Local4 }, { "local5", Local5 }, { "local6", Local6 }, { "local7", Local7 },
    { NULL, (Level) - 1  }
  };
  int i = 0;

  while(values [i].stringLevel != NULL) {
    if(strcasecmp(stringLevel, values [i].stringLevel) == 0)
      break;

    i ++;
  }

  if(values [i].stringLevel == NULL) {
    std::string msg = "Level not registered: '";
    msg += stringLevel;
    msg += "'. Possible values: ";

    for(i = 0; values [i].stringLevel != NULL; i ++) {
      msg += values [i].stringLevel;
      msg += ' ';
    }

    throw RuntimeException(msg, ANNA_FILE_LOCATION);
  }

  return values [i].level;
}

const char* Logger::asString(const Level level)
throw() {
  static const char* levels [] = {
    "Emergency", "Alert", "Critical", "Error", "Warning", "Notice", "Information", "Debug"
  };
  static struct {
    const char* stringLevel;
    const Level level;
  } values [] = {
    { "Local0", Local0 }, { "Local1", Local1 }, { "Local2", Local2 }, { "Local3", Local3 },
    { "Local4", Local4 }, { "Local5", Local5 }, { "Local6", Local6 }, { "Local7", Local7 },
    { NULL, (Level) - 1  }
  };
  const char* result = NULL;

  if(level >= Emergency && level <= Debug)
    result = levels [level];
  else {
    for(int i = 0; values [i].stringLevel != NULL; i ++) {
      if(level == values [i].level) {
        result = values [i].stringLevel;
        break;
      }
    }

    if(result == NULL)
      result = "Other";
  }

  return result;
}

Logger::Writer::Writer(const int bufferSize) {
  try {
    a_dataBlock = new DataBlock(true);
    a_dataBlock->allocate(bufferSize);
  } catch(Exception&) {
  }
}

Logger::Writer::Writer() {
  try {
    a_dataBlock = new DataBlock(true);
    a_dataBlock->allocate(1024);
  } catch(Exception&) {
  }
}

Logger::Writer::~Writer() {
  Logger::st_writer = NULL;
}

