// ANNA - Anna is Not Nothingness Anymore                                                         //
//                                                                                                //
// (c) Copyright 2005-2015 Eduardo Ramos Testillano & Francisco Ruiz Rayo                         //
//                                                                                                //
// See project site at http://redmine.teslayout.com/projects/anna-suite                           //
// See accompanying file LICENSE or copy at http://www.teslayout.com/projects/public/anna.LICENSE //


// Standard
#include <algorithm>

// Project
#include <anna/diameter.comm/ApplicationMessageOamModule.hpp>
#include <anna/core/mt/Guard.hpp>
#include <anna/core/functions.hpp>
#include <anna/core/oam/CounterScope.hpp>
#include <anna/diameter/stack/Engine.hpp>


anna_assign_enum(anna::diameter::comm::ApplicationMessageOamModule::Counter) = { \
    "Request_SentOK_AsClient", \
    "Request_SentNOK_AsClient", \
    "Answer_SentOK_AsClient", \
    "Answer_SentNOK_AsClient", \
    "Request_Received_AsClient", \
    "Answer_Received_AsClient", \
    "Answer_UnknownReceived_AsClient", \

    "Request_SentOK_AsServer", \
    "Request_SentNOK_AsServer", \
    "Answer_SentOK_AsServer", \
    "Answer_SentNOK_AsServer", \
    "Request_Received_AsServer", \
    "Answer_Received_AsServer", \
    "Answer_UnknownReceived_AsServer", \
    NULL /* list end indicator */
};

void anna::diameter::comm::ApplicationMessageOamModule::createStackCounterScope(int scopeId, unsigned int stackId) throw(anna::RuntimeException) {

  initializeCounterScope(scopeId, anna::functions::asString("Application Message Events for stack id %lu", stackId));

  // Better to be enabled by application in order to be more conscious of the native disabled nature of this special oam module
  //if (!countersEnabled()) enableCounters();
  //enableAlarms(); not yet implemented

  a_stackMap[stackId] = scopeId;
}

void anna::diameter::comm::ApplicationMessageOamModule::count (int messageCode, int resultCode, unsigned int stackId, const int & type, const int & amount) throw(anna::RuntimeException) {

  // Optimization:
  // Checkings
  if(!countersEnabled()) {
    LOGDEBUG
    (
      std::string msg = "Count operation ignored over module '";
      msg += getName();
      msg += "': counters are disabled";
      anna::Logger::debug(msg, ANNA_FILE_LOCATION);
    );
    return;
  }

  anna::Guard guard(a_mutex, "ApplicationMessageOamModule::count"); // counter scope switch

  int scopeId = monoStackScopeId();
  if (scopeId == -1) {
    std::map<unsigned int /* stack id */, int /* scope id */>::const_iterator stackMap_it = a_stackMap.find(stackId);

    if (stackMap_it == a_stackMap.end()) {
      LOGDEBUG(anna::Logger::debug(anna::functions::asString("Unregistered stack id %lu", stackId), ANNA_FILE_LOCATION));
      return;
    }

    scopeId = stackMap_it->second;
  }

  // Select counter scope 
  setActiveCounterScope(scopeId);

  // Build event id: <message code>_<result code>
  std::string eventId = anna::functions::asString("%d_%d", messageCode, resultCode); 

  std::map<std::string /* event id */, int /* base offset */>::const_iterator eventMap_it = a_eventMap.find(eventId);
  int baseOffset;
     
  if (eventMap_it == a_eventMap.end()) {
    int capacity = anna::oam::CounterScope::MaxCounter / getCounterTypes();
    if (a_eventMap.size() > capacity) {
      LOGDEBUG(anna::Logger::debug(anna::functions::asString("No more holes to register new application message counters in the scope (up to %d message codes)", capacity), ANNA_FILE_LOCATION));
      return;
    }
    baseOffset = getCounterTypes() * a_eventMap.size(); // N counter types for each message code / rc
    a_eventMap[eventId] = baseOffset;

    // Counter name:
    std::string counterNamePrefix = anna::functions::asString("ApplicationMessageCode_%d", messageCode); // default
    anna::diameter::stack::Engine &stackEngine = anna::diameter::stack::Engine::instantiate();
    const anna::diameter::stack::Dictionary *dictionary = stackEngine.getDictionary(stackId);
    if (dictionary) {
      anna::diameter::CommandId cidR(messageCode, true /* request */);
      anna::diameter::CommandId cidA(messageCode, false /* answer */);
      const anna::diameter::stack::Command *commandR = dictionary->getCommand(cidR);
      const anna::diameter::stack::Command *commandA = dictionary->getCommand(cidA);
      if (commandR && commandA) {
        std::string string1 = commandR->getName();
        std::string string2 = commandA->getName();
        std::string intersection; // non-standard names will be also intersected: XXR and XXA gives XX
        std::set_intersection(string1.begin(), string1.end(), string2.begin(), string2.end(), std::back_inserter(intersection));
        counterNamePrefix = intersection;
        // Special case:
        if (counterNamePrefix == "") counterNamePrefix = string1 + "#" + string2;
      }
    } 

    if (counterNamePrefix[counterNamePrefix.size() - 1] != '-') counterNamePrefix += "-";
    std::string counterName;
    /*
    for (int offset = 0; offset < getCounterTypes(); offset++) {
      counterName = counterNamePrefix + getDefaultInternalCounterDescription(offset);
      if (resultCode != -1) counterName += anna::functions::asString("-ResultCode_%d", resultCode);
      registerCounter(baseOffset + offset, counterName, baseOffset + offset);
    }
    */
    // Register only affected one:
    int offset = baseOffset + type;
    counterName = counterNamePrefix + getDefaultInternalCounterDescription(type);
    if (resultCode != -1) counterName += anna::functions::asString("-ResultCode_%d", resultCode);
    registerCounter(offset, counterName, offset);
    
  }
  else {
    baseOffset = eventMap_it->second;
  }

  // Count
  Module::count(baseOffset + type, amount);
}

